import json
from typing import List, Dict, Any, Union

import requests
from tenacity import retry, stop_after_attempt, wait_random

from smhi_open_data.enums import Parameter
from smhi_open_data.utils import try_parse_float


class SMHIOpenDataClient:
    _base_url = 'https://opendata-download-metobs.smhi.se'

    def __init__(self, version: str = '1.0'):
        self.version = version

    @property
    def base_url(self):
        return f"{self._base_url}/api/version/{self.version}"

    @retry(stop=stop_after_attempt(10), wait=wait_random(min=0.01, max=0.10))
    def _query(self, service: str, params: Dict[str, Any] = None, **kwargs):
        # Create a session
        s = requests.Session()

        # Create adapters with the retry logic for each
        http = requests.adapters.HTTPAdapter(max_retries=30)
        https = requests.adapters.HTTPAdapter(max_retries=30)

        # Replace the session's original adapters
        s.mount('http://', http)
        s.mount('https://', https)

        # Start using the session
        res = s.get(url=f"{self.base_url}/{service}", stream=True)

        if res.encoding is None:
            res.encoding = 'utf-8'

        data = []
        for line in res.iter_lines(decode_unicode=True):
            if line:
                data.append(line)
        return json.loads(''.join(data))

    def get_parameter_stations(self, parameter: Parameter) -> List[Dict[str, Any]]:
        """Get parameter stations.

        Returns:
            List[Dict[str, Any]]: List of SMHI stations where parameter is available.
        """
        res = self._query(
            service=f"parameter/{parameter.value}.json")
        return res['station']

    def get_station_parameters(self,
                               station_id: int,
                               parameter_set: List[Parameter] = list(Parameter)) -> List[Parameter]:
        parameters = set([])
        for parameter in parameter_set:
            for station in self.get_parameter_stations(parameter=parameter):
                if station['id'] == station_id:
                    parameters.add(parameter)
        return list(parameters)

    def get_stations(self) -> List[Dict[str, Any]]:
        """Get SMHI stations. NOTE: Performs multiple http requests.

        NOTE: This performs multiple requests and might take long time to run.

        Returns:
            List[Dict[str, Any]]: List of SMHI stations.
        """
        stations, station_ids = [], set([])
        for param in Parameter:
            stations = self.get_parameter_stations(parameter=param)
            for station in stations:
                if station['id'] not in station_ids:
                    station_ids.add(station['id'])
                    stations.append(station)
        return stations

    def get_latest_observations(self, parameter: Parameter) -> List[Dict[str, Any]]:
        """Get latest raw SMHI observations from available stations.

        Args:
            parameter_id (Parameter): Returns observations for a specific parameter.

        Returns:
            List[Dict[str, Any]]: List of raw SMHI observations from available stations.
        """
        # Check if parameter has station set
        res = self._query(
            service=f"parameter/{parameter.value}/station-set.json"
        )
        station_set = res['stationSet']

        if station_set is None:
            # TODO: Get values from stations instead
            raise NotImplementedError(f"Not implemented for parameter: {parameter}")

        # Check if station set has any periods
        res = self._query(
            service=f"parameter/{parameter.value}/station-set/all.json"
        )
        periods = res['period']
        if periods is None or len(periods) == 0:
            raise NotImplementedError(f"Not implemented for parameter: {parameter}")

        # Get period key
        period_key = periods[0]['key']
        if period_key != 'latest-hour':
            raise NotImplementedError(f"Not implemented for parameter: {parameter} and period: {period_key}")

        # Get period data
        res = self._query(
            service=f"parameter/{parameter.value}/station-set/all/period/latest-hour/data.json"
        )

        values = []
        for x in res['station']:
            value_list = x['value']
            if value_list is None:
                continue

            for value in value_list:
                values.append({
                    'parameter_id': parameter.value,
                    'timestamp': value['date'],
                    'value': try_parse_float(value['value']),
                    'station': x['key'],
                })

        return values

    def list_parameters(self) -> List[Dict[str, Union[str, Parameter]]]:
        """List available observation parameters.

        Returns:
            List[Dict[str, Union[str, Parameter]]]: List of dictionaries
                containing information about each available observations
                parameter.
        """
        return [
            {
                'name': parameter.name,
                'value': parameter.value,
                'enum': parameter,
            }
            for parameter in Parameter
        ]

    @staticmethod
    def get_parameter(parameter_id: int) -> Parameter:
        """Get parameter enum from SMHI parameter id.

        Args:
            parameter_id (int): Parameter id found on SMHI API documentation.

        Returns:
            Parameter: Parameter enum object.
        """
        return Parameter(parameter_id)



from datetime import datetime
from typing import Union


EPOCH = datetime.utcfromtimestamp(0)


def date2microseconds(date: datetime) -> int:
    return int((date - EPOCH).total_seconds() * 1000000.0)


def microseconds2date(microseconds: float) -> datetime:
    return datetime.utcfromtimestamp(microseconds / 1000000)


def try_parse_float(x: Union[str, float, int]) -> Union[float, str]:
    try:
        return float(x)
    except ValueError:
        return x



from smhi_open_data.client import SMHIOpenDataClient
from smhi_open_data.enums import Parameter
from smhi_open_data.utils import microseconds2date, date2microseconds


__all__ = [
    'SMHIOpenDataClient',
    'Parameter',
    'microseconds2date',
    'date2microseconds',
]



from enum import Enum


# See https://opendata.smhi.se/apidocs/metobs/parameter.html
class Parameter(Enum):
    TemperaturePast1h = 1               # Lufttemperatur                        Momentanvärde, 1 gång/tim
    TemperaturePast24h = 2              # Lufttemperatur                        Medelvärde 1 dygn, 1 gång/dygn, kl 00
    WindDirection = 3                   # Vindriktning                          Medelvärde 10 min, 1 gång/tim
    WindSpeed = 4                       # Vindhastighet	                        Medelvärde 10 min, 1 gång/tim
    PrecipPast24hAt06 = 5               # Nederbördsmängd                       Summa 1 dygn, 1 gång/dygn, kl 06
    Humidity = 6                        # Relativ Luftfuktighet	                Momentanvärde, 1 gång/tim
    PrecipPast1h = 7                    # Nederbördsmängd                       Summa 1 timme, 1 gång/tim
    SnowDepthPast24h = 8                # Snödjup                               Momentanvärde, 1 gång/dygn, kl 06
    Pressure = 9                        # Lufttryck reducerat havsytans nivå    Vid havsytans nivå, momentanvärde, 1 gång/tim
    SunLast1h = 10                      # Solskenstid                           Summa 1 timme, 1 gång/tim
    RadiaGlob = 11                      # Global Irradians (svenska stationer)  Medelvärde 1 timme, 1 gång/tim
    Visibility = 12                     # Sikt                                  Momentanvärde, 1 gång/tim
    CurrentWeather = 13                 # Rådande väder                         Momentanvärde, 1 gång/tim resp 8 gånger/dygn
    PrecipPast15m = 14                  # Nederbördsmängd                       Summa 15 min, 4 gånger/tim
    PrecipMaxPast15m = 15               # Nederbördsintensitet                  Max under 15 min, 4 gånger/tim
    CloudCover = 16                     # Total molnmängd                       Momentanvärde, 1 gång/tim
    PrecipPast12h = 17                  # Nederbörd                             2 gånger/dygn, kl 06 och 18
    PrecipPast24hAt18 = 18              # Nederbörd                             1 gång/dygn, kl 18
    TemperatureMinPast24h = 19          # Lufttemperatur                        Min, 1 gång per dygn
    TemperatureMaxPast24h = 20          # Lufttemperatur                        Max, 1 gång per dygn
    WindSpeedTown = 21                  # Byvind                                Max, 1 gång/tim
    TemperatureMeanPastMonth = 22       # Lufttemperatur                        Medel, 1 gång per månad
    PrecipPastMonth = 23                # Nederbördsmängd                       Summa, 1 gång per månad
    LongwaveIrradians = 24              # Långvågs-Irradians                    Långvågsstrålning, medel 1 timme, varje timme
    WindSpeedMaxMeanPast3h = 25         # Max av MedelVindhastighet             Maximum av medelvärde 10 min, under 3 timmar, ...
    TemperatureMinPast12h = 26          # Lufttemperatur                        Min, 2 gånger per dygn, kl 06 och 18
    TemperatureMaxPast12h = 27	        # Lufttemperatur                        Max, 2 gånger per dygn, kl 06 och 18
    CloudLayerLowest = 28               # Molnbas                               Lägsta molnlager, momentanvärde, 1 gång/tim
    CloudAmountLowest = 29              # Molnmängd                             Lägsta molnlager, momentanvärde, 1 gång/tim
    CloudLayerOther = 30                # Molnbas                               Andra molnlager, momentanvärde, 1 gång/tim
    CloudAmountOther = 31               # Molnmängd                             Andra molnlager, momentanvärde, 1 gång/tim
    CloudLayer3rd = 32                  # Molnbas                               Tredje molnlager, momentanvärde, 1 gång/tim
    CloudAmount3rd = 33                 # Molnmängd                             Tredje molnlager, momentanvärde, 1 gång/tim
    CloudLayer4th = 34                  # Molnbas                               Fjärde molnlager, momentanvärde, 1 gång/tim
    CloudAmount4th = 35                 # Molnmängd                             Fjärde molnlager, momentanvärde, 1 gång/tim
    CloudStorageLowest = 36             # Molnbas                               Lägsta molnbas, momentanvärde, 1 gång/tim
    CloudStorageLowestMin = 37          # Molnbas                               Lägsta molnbas, min under 15 min, 1 gång/tim
    PrecipIntensityMaxMeanPast15m = 38  # Nederbördsintensitet                  Max av medel under 15 min, 4 gånger/tim
    TemperatureDew = 39                 # Daggpunktstemperatur                  Momentanvärde, 1 gång/tim
    GroundCondition = 40                # Markens tillstånd                     Momentanvärde, 1 gång/dygn, kl 06



import unittest
import warnings

from smhi_open_data import SMHIOpenDataClient, Parameter, microseconds2date


warnings.simplefilter('ignore', ResourceWarning)


class TestClient(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.client = SMHIOpenDataClient()

    def test_stations(self):
        stations = self.client.get_stations()
        self.assertIsInstance(stations, list, 'Did not return a list of stations')
        self.assertGreater(len(stations), 0, 'Could not find any stations')

    def test_observations(self):
        observations = self.client.get_latest_observations(parameter=Parameter.TemperaturePast1h)
        self.assertIsInstance(observations, list, 'Did not return a list of observations')
        self.assertGreater(len(observations), 0, 'Could not find any observations')

    def test_list_parameters(self):
        parameters = self.client.list_parameters()
        self.assertTrue(
            all(
                isinstance(Parameter(parameter['value']), Parameter)
                for parameter in parameters
            ),
            'Returned a wrong parameter')

    def test_get_parameter(self):
        parameter = self.client.get_parameter(parameter_id=1)
        self.assertIsInstance(parameter, Parameter, 'Returned value was not a Parameter object.')
        self.assertEqual(parameter, Parameter.TemperaturePast1h, 'Wrong parameter returned.')

    def test_get_parameter_stations(self):
        stations = self.client.get_parameter_stations(parameter=Parameter.TemperaturePast1h)
        self.assertIsInstance(stations, list, 'Returned stations was not a list.')
        self.assertGreater(len(stations), 0, 'No stations returned.')
        self.assertIsInstance(stations[0], dict, 'Returned stations were not dicts.')

    def test_get_station_parameters(self):
        parameters = self.client.get_station_parameters(station_id=173010)
        self.assertIsInstance(parameters, list, 'Returned parameters was not a list.')
        self.assertGreater(len(parameters), 0, 'No parameters returned.')
        self.assertIsInstance(parameters[0], Parameter, 'Returned parameters were not parameter nums.')


if __name__ == '__main__':
    unittest.main()
